@@ -1,4 +1,4 @@
-/* $Id: files.c,v 1.11 2004/01/09 23:04:55 dolorous Exp $ */
+/* $Id: files.c,v 1.25 2004/03/19 21:46:34 dolorous Exp $ */
 /**************************************************************************
  *   files.c                                                              *
  *                                                                        *
@@ -62,7 +62,8 @@ void load_file(int update)
 
 #ifdef ENABLE_COLOR
     update_color();
-    edit_refresh();
+    if (ISSET(COLOR_SYNTAX))
+	edit_refresh();
 #endif
 }
 
@@ -74,7 +75,6 @@ void new_file(void)
     fileage->data[0] = '\0';
     filebot = fileage;
     edittop = fileage;
-    editbot = fileage;
     current = fileage;
     current_x = 0;
     totlines = 1;
@@ -100,7 +100,8 @@ void new_file(void)
 
 #ifdef ENABLE_COLOR
     update_color();
-    edit_refresh();
+    if (ISSET(COLOR_SYNTAX))
+	edit_refresh();
 #endif
 }
 
@@ -404,7 +405,7 @@ char *get_next_filename(const char *name)
     buf = charalloc(strlen(name) + num_of_digits(INT_MAX) + 2);
     strcpy(buf, name);
 
-    while (1) {
+    while (TRUE) {
 
 	if (stat(buf, &fs) == -1)
 	    return buf;
@@ -446,20 +447,20 @@ int do_insertfile(int loading_file)
 #endif
 
 #ifndef DISABLE_OPERATINGDIR
-    if (operating_dir != NULL && strcmp(operating_dir, ".") != 0)
+    if (operating_dir != NULL && strcmp(operating_dir, "."))
 #ifdef ENABLE_MULTIBUFFER 
 	if (ISSET(MULTIBUFFER))
-	    i = statusq(1, insertfile_list, inspath,
+	    i = statusq(TRUE, insertfile_list, inspath,
 #ifndef NANO_SMALL
-		0,
+		NULL,
 #endif
 		_("File to insert into new buffer [from %s] "),
 		operating_dir);
 	else
 #endif
-	    i = statusq(1, insertfile_list, inspath,
+	    i = statusq(TRUE, insertfile_list, inspath,
 #ifndef NANO_SMALL
-		0,
+		NULL,
 #endif
 		_("File to insert [from %s] "),
 		operating_dir);
@@ -468,16 +469,16 @@ int do_insertfile(int loading_file)
 #endif
 #ifdef ENABLE_MULTIBUFFER 
 	if (ISSET(MULTIBUFFER))
-	    i = statusq(1, insertfile_list, inspath,
+	    i = statusq(TRUE, insertfile_list, inspath,
 #ifndef NANO_SMALL
-		0,
+		NULL,
 #endif
 		_("File to insert into new buffer [from ./] "));
 	else
 #endif /* ENABLE_MULTIBUFFER */
-	    i = statusq(1, insertfile_list, inspath,
+	    i = statusq(TRUE, insertfile_list, inspath,
 #ifndef NANO_SMALL
-		0,
+		NULL,
 #endif
 		_("File to insert [from ./] "));
 
@@ -499,8 +500,12 @@ int do_insertfile(int loading_file)
 #endif /* ENABLE_MULTIBUFFER */
 
 	if (i == NANO_EXTCMD_KEY) {
-	    int ts = statusq(TRUE, extcmd_list, "", NULL, 
+	    char *ans = mallocstrcpy(NULL, answer);
+	    int ts = statusq(TRUE, extcmd_list, ans, NULL, 
 		_("Command to execute"));
+
+	    free(ans);
+
 	    if (ts  == -1 || answer == NULL || answer[0] == '\0') {
 		statusbar(_("Cancelled"));
 		display_main_list();
@@ -522,7 +527,11 @@ int do_insertfile(int loading_file)
 #endif
 
 #ifndef DISABLE_OPERATINGDIR
-	if (i != NANO_EXTCMD_KEY && check_operating_dir(answer, 0) != 0) {
+	if (
+#ifndef NANO_SMALL
+		i != NANO_EXTCMD_KEY &&
+#endif
+		check_operating_dir(answer, 0) != 0) {
 	    statusbar(_("Can't insert file from outside of %s"),
 			operating_dir);
 	    return 0;
@@ -582,9 +591,6 @@ int do_insertfile(int loading_file)
 #endif
 	    set_modified();
 
-	/* Here we want to rebuild the edit window */
-	fix_editbot();
-
 #ifdef ENABLE_MULTIBUFFER
 	/* If we've loaded another file, update the titlebar's contents */
 	if (loading_file) {
@@ -608,10 +614,7 @@ int do_insertfile(int loading_file)
 #endif
 
 	/* If we've gone off the bottom, recenter; otherwise, just redraw */
-	if (current->lineno > editbot->lineno)
-	    edit_update(current, CENTER);
-	else
-	    edit_refresh();
+	edit_refresh();
 
     } else {
 	statusbar(_("Cancelled"));
@@ -775,17 +778,24 @@ int add_open_file(int update)
     /* save current line number */
     open_files->file_lineno = current->lineno;
 
+    /* start with default modification status: unmodified (and marking
+       status, if available: unmarked) */
+    open_files->file_flags = 0;
+
     /* if we're updating, save current modification status (and marking
        status, if available) */
     if (update) {
 #ifndef NANO_SMALL
-	open_files->file_flags = (MODIFIED & ISSET(MODIFIED)) | (MARK_ISSET & ISSET(MARK_ISSET));
+	if (ISSET(MODIFIED))
+	    open_files->file_flags |= MODIFIED;
 	if (ISSET(MARK_ISSET)) {
 	    open_files->file_mark_beginbuf = mark_beginbuf;
 	    open_files->file_mark_beginx = mark_beginx;
+	    open_files->file_flags |= MARK_ISSET;
 	}
 #else
-	open_files->file_flags = (MODIFIED & ISSET(MODIFIED));
+	if (ISSET(MODIFIED))
+	    open_files->file_flags |= MODIFIED;
 #endif
     }
 
@@ -1017,7 +1027,7 @@ int close_open_file(void)
 }
 #endif /* MULTIBUFFER */
 
-#if !defined(DISABLE_SPELLER) || !defined(DISABLE_OPERATINGDIR)
+#if !defined(DISABLE_SPELLER) || !defined(DISABLE_OPERATINGDIR) || !defined(NANO_SMALL)
 /*
  * When passed "[relative path]" or "[relative path][filename]" in
  * origpath, return "[full path]" or "[full path][filename]" on success,
@@ -1260,10 +1270,11 @@ void init_operating_dir(void)
 
     if (operating_dir == NULL)
 	return;
+
     full_operating_dir = get_full_path(operating_dir);
 
     /* If get_full_path() failed or the operating directory is
-       inaccessible, unset operating_dir. */
+     * inaccessible, unset operating_dir. */
     if (full_operating_dir == NULL || chdir(full_operating_dir) == -1) {
 	free(full_operating_dir);
 	full_operating_dir = NULL;
@@ -1290,6 +1301,7 @@ int check_operating_dir(const char *currpath, int allow_tabcomp)
     /* If no operating directory is set, don't bother doing anything. */
     if (operating_dir == NULL)
 	return 0;
+
     assert(full_operating_dir != NULL);
 
     fullpath = get_full_path(currpath);
@@ -1322,6 +1334,30 @@ int check_operating_dir(const char *currpath, int allow_tabcomp)
 }
 #endif
 
+#ifndef NANO_SMALL
+void init_backup_dir(void)
+{
+    char *full_backup_dir;
+
+    if (backup_dir == NULL)
+	return;
+
+    full_backup_dir = get_full_path(backup_dir);
+
+    /* If get_full_path() failed or the backup directory is
+     * inaccessible, unset backup_dir. */
+    if (full_backup_dir == NULL ||
+	full_backup_dir[strlen(full_backup_dir) - 1] != '/') {
+	free(full_backup_dir);
+	free(backup_dir);
+	backup_dir = NULL;
+    } else {
+	free(backup_dir);
+	backup_dir = full_backup_dir;
+    }
+}
+#endif
+
 /* Read from inn, write to out.  We assume inn is opened for reading,
  * and out for writing.  We return 0 on success, -1 on read error, -2 on
  * write error. */
@@ -1454,8 +1490,38 @@ int write_file(const char *name, int tmp, int append, int nonamechange)
 	    goto cleanup_and_exit;
 	}
 
-	backupname = charalloc(strlen(realname) + 2);
-	sprintf(backupname, "%s~", realname);
+	/* If backup_dir is set, we set backupname to
+	 * backup_dir/backupname~, where backupnae is the canonicalized
+	 * absolute pathname of realname with every '/' replaced with a
+	 * '!'.  This means that /home/foo/file is backed up in
+	 * backup_dir/!home!foo!file~. */
+	if (backup_dir != NULL) {
+	    char *canon_realname = get_full_path(realname);
+	    size_t i;
+
+	    if (canon_realname == NULL)
+		/* If get_full_path() failed, we don't have a
+		 * canonicalized absolute pathname, so just use the
+		 * filename portion of the pathname.  We use tail() so
+		 * that e.g. ../backupname will be backed up in
+		 * backupdir/backupname~ instead of
+		 * backupdir/../backupname~. */
+		canon_realname = mallocstrcpy(NULL, tail(realname));
+	    else {
+		for (i = 0; canon_realname[i] != '\0'; i++) {
+		    if (canon_realname[i] == '/')
+			canon_realname[i] = '!';
+		}
+	    }
+
+	    backupname = charalloc(strlen(backup_dir) +
+		strlen(canon_realname) + 2);
+	    sprintf(backupname, "%s%s~", backup_dir, canon_realname);
+	    free(canon_realname);
+	} else {
+	    backupname = charalloc(strlen(realname) + 2);
+	    sprintf(backupname, "%s~", realname);
+	}
 
 	/* Open the destination backup file.  Before we write to it, we
 	 * set its permissions, so no unauthorized person can read it as
@@ -1647,7 +1713,8 @@ int write_file(const char *name, int tmp, int append, int nonamechange)
 	    filename = mallocstrcpy(filename, realname);
 #ifdef ENABLE_COLOR
 	    update_color();
-	    edit_refresh();
+	    if (ISSET(COLOR_SYNTAX))
+		edit_refresh();
 #endif
 	}
 
@@ -1669,35 +1736,113 @@ int write_file(const char *name, int tmp, int append, int nonamechange)
     return retval;
 }
 
-int do_writeout(const char *path, int exiting, int append)
+#ifndef NANO_SMALL
+/* Write a marked selection from a file out.  First, set fileage and
+ * filebot as the top and bottom of the mark, respectively.  Then call
+ * write_file() with the values of name, temp, append, and nonamechange.
+ * Finally, set fileage and filebot back to their old values and
+ * return. */
+int write_marked(const char *name, int tmp, int append, int
+	nonamechange)
 {
-    int i = 0;
+    int retval = -1;
+    filestruct *fileagebak = fileage;
+    filestruct *filebotbak = filebot;
+    int oldmod = ISSET(MODIFIED);
+	/* write_file() unsets the MODIFIED flag. */
+    size_t topx;
+	/* The column of the beginning of the mark. */
+    char origchar;
+	/* We replace the character at the end of the mark with '\0'.
+	 * We save the original character, to restore it. */
+    char *origcharloc;
+	/* The location of the character we nulled. */
+
+    if (!ISSET(MARK_ISSET))
+	return -1;
+
+    /* Set fileage as the top of the mark, and filebot as the bottom. */
+    if (current->lineno > mark_beginbuf->lineno ||
+		(current->lineno == mark_beginbuf->lineno &&
+		current_x > mark_beginx)) {
+	fileage = mark_beginbuf;
+	topx = mark_beginx;
+	filebot = current;
+	origcharloc = current->data + current_x;
+    } else {
+	fileage = current;
+	topx = current_x;
+	filebot = mark_beginbuf;
+	origcharloc = mark_beginbuf->data + mark_beginx;
+    }
+    origchar = *origcharloc;
+    *origcharloc = '\0';
+    fileage->data += topx;
+
+    /* If the line at filebot is blank, treat it as the magicline and
+     * hence the end of the file.  Otherwise, treat the line after
+     * filebot as the end of the file. */
+    if (filebot->data[0] != '\0' && filebot->next != NULL)
+	filebot = filebot->next;
+
+    retval = write_file(name, tmp, append, nonamechange);
+
+    /* Now restore everything. */
+    fileage->data -= topx;
+    *origcharloc = origchar;
+    fileage = fileagebak;
+    filebot = filebotbak;
+    if (oldmod)
+	set_modified();
+
+    return retval;
+}
+#endif /* !NANO_SMALL */
+
+int do_writeout(int exiting)
+{
+    int i;
+    int append = 0;
 #ifdef NANO_EXTRA
-    static int did_cred = 0;
+    static int did_cred = FALSE;
 #endif
 
 #if !defined(DISABLE_BROWSER) || !defined(DISABLE_MOUSE)
     currshortcut = writefile_list;
 #endif
 
-    answer = mallocstrcpy(answer, path);
-
     if (exiting && ISSET(TEMP_OPT)) {
+	i = -1;
 	if (filename[0] != '\0') {
-	    i = write_file(answer, 0, 0, 0);
-	    display_main_list();
-	    return i;
-	} else {
+	    i = write_file(filename, FALSE, 0, FALSE);
+	    if (i == 1) {
+		/* Write succeeded. */
+		display_main_list();
+		return 1;
+	    }
+	}
+
+	/* No filename or the write above failed. */
+	if (i == -1) {
 	    UNSET(TEMP_OPT);
 	    do_exit();
 
-	    /* They cancelled, abort quit */
+	    /* They cancelled; abort quit. */
 	    return -1;
 	}
     }
 
-    while (1) {
 #ifndef NANO_SMALL
+    if (ISSET(MARK_ISSET) && !exiting)
+	answer = mallocstrcpy(answer, "");
+    else
+#endif
+	answer = mallocstrcpy(answer, filename);
+
+    while (TRUE) {
+	const char *msg;
+#ifndef NANO_SMALL
+	char *ans = mallocstrcpy(NULL, answer);
 	const char *formatstr, *backupstr;
 
 	if (ISSET(MAC_FILE))
@@ -1712,44 +1857,39 @@ int do_writeout(const char *path, int exiting, int append)
 	else
 	   backupstr = "";
 
-	/* Be nice to the translation folks */
+	/* Be nice to the translation folks. */
 	if (ISSET(MARK_ISSET) && !exiting) {
 	    if (append == 2)
-		i = statusq(1, writefile_list, "", 0,
-		    "%s%s%s", _("Prepend Selection to File"), formatstr, backupstr);
+		msg = _("Prepend Selection to File");
 	    else if (append == 1)
-		i = statusq(1, writefile_list, "", 0,
-		    "%s%s%s", _("Append Selection to File"), formatstr, backupstr);
+		msg = _("Append Selection to File");
 	    else
-		i = statusq(1, writefile_list, "", 0,
-		    "%s%s%s", _("Write Selection to File"), formatstr, backupstr);
-	} else {
-	    if (append == 2)
-		i = statusq(1, writefile_list, answer, 0,
-		    "%s%s%s", _("File Name to Prepend to"), formatstr, backupstr);
-	    else if (append == 1)
-		i = statusq(1, writefile_list, answer, 0,
-		    "%s%s%s", _("File Name to Append to"), formatstr, backupstr);
-	    else
-		i = statusq(1, writefile_list, answer, 0,
-		    "%s%s%s", _("File Name to Write"), formatstr, backupstr);
-	}
-#else
-	if (append == 2)
-	    i = statusq(1, writefile_list, answer,
-		"%s", _("File Name to Prepend to"));
-	else if (append == 1)
-	    i = statusq(1, writefile_list, answer,
-		"%s", _("File Name to Append to"));
-	else
-	    i = statusq(1, writefile_list, answer,
-		"%s", _("File Name to Write"));
+		msg = _("Write Selection to File");
+	} else
 #endif /* !NANO_SMALL */
+	if (append == 2)
+	    msg = _("File Name to Prepend to");
+	else if (append == 1)
+	    msg = _("File Name to Append to");
+	else
+	    msg = _("File Name to Write");
+
+	i = statusq(TRUE, writefile_list,
+#ifndef NANO_SMALL
+		ans, NULL, "%s%s%s", msg, formatstr, backupstr
+#else
+		filename, "%s", msg
+#endif
+		);
+
+#ifndef NANO_SMALL
+	free(ans);
+#endif
 
 	if (i == -1) {
 	    statusbar(_("Cancelled"));
 	    display_main_list();
-	    return 0;
+	    return -1;
 	}
 
 #ifndef DISABLE_BROWSER
@@ -1793,7 +1933,7 @@ int do_writeout(const char *path, int exiting, int append)
 	if (exiting && !ISSET(TEMP_OPT) && !strcasecmp(answer, "zzy")
 		&& !did_cred) {
 	    do_credits();
-	    did_cred = 1;
+	    did_cred = TRUE;
 	    return -1;
 	}
 #endif
@@ -1801,7 +1941,7 @@ int do_writeout(const char *path, int exiting, int append)
 	    struct stat st;
 
 	    if (!stat(answer, &st)) {
-		i = do_yesno(0, _("File exists, OVERWRITE ?"));
+		i = do_yesno(FALSE, _("File exists, OVERWRITE ?"));
 		if (i == 0 || i == -1)
 		    continue;
 	    } else if (filename[0] != '\0'
@@ -1809,7 +1949,7 @@ int do_writeout(const char *path, int exiting, int append)
 		&& (!ISSET(MARK_ISSET) || exiting)
 #endif
 		) {
-		i = do_yesno(0, _("Save file under DIFFERENT NAME ?"));
+		i = do_yesno(FALSE, _("Save file under DIFFERENT NAME ?"));
 		if (i == 0 || i == -1)
 		    continue;
 	    }
@@ -1818,70 +1958,26 @@ int do_writeout(const char *path, int exiting, int append)
 #ifndef NANO_SMALL
 	/* Here's where we allow the selected text to be written to
 	 * a separate file. */
-	if (ISSET(MARK_ISSET) && !exiting) {
-	    filestruct *fileagebak = fileage;
-	    filestruct *filebotbak = filebot;
-	    int oldmod = ISSET(MODIFIED);
-		/* write_file() unsets the MODIFIED flag. */
-	    size_t topx;
-		/* The column of the beginning of the mark. */
-	    char origchar;
-		/* We replace the character at the end of the mark with
-		 * '\0'.  We save the original character, to restore
-		 * it. */
-	    char *origcharloc;
-		/* The location of the character we nulled. */
-
-	    /* Set fileage as the top of the mark, and filebot as the
-	     * bottom. */
-	    if (current->lineno > mark_beginbuf->lineno ||
-		(current->lineno == mark_beginbuf->lineno &&
-		current_x > mark_beginx)) {
-		fileage = mark_beginbuf;
-		topx = mark_beginx;
-		filebot = current;
-		origcharloc = current->data + current_x;
-	    } else {
-		fileage = current;
-		topx = current_x;
-		filebot = mark_beginbuf;
-		origcharloc = mark_beginbuf->data + mark_beginx;
-	    }
-	    origchar = *origcharloc;
-	    *origcharloc = '\0';
-	    fileage->data += topx;
-	    /* If the line at filebot is blank, treat it as the
-	     * magicline and hence the end of the file.  Otherwise,
-	     * treat the line after filebot as the end of the file. */
-	    if (filebot->data[0] != '\0' && filebot->next != NULL)
-		filebot = filebot->next;
-	    i = write_file(answer, 0, append, 1);
-
-	    /* Now restore everything. */
-	    fileage->data -= topx;
-	    *origcharloc = origchar;
-	    fileage = fileagebak;
-	    filebot = filebotbak;
-	    if (oldmod)
-		set_modified();
-	} else
+	if (ISSET(MARK_ISSET) && !exiting)
+	    i = write_marked(answer, FALSE, append, FALSE);
+	else
 #endif /* !NANO_SMALL */
-	    i = write_file(answer, 0, append, 0);
+	    i = write_file(answer, FALSE, append, FALSE);
 
 #ifdef ENABLE_MULTIBUFFER
 	/* If we're not about to exit, update the current entry in
-	   the open_files structure. */
+	 * the open_files structure. */
 	if (!exiting)
 	    add_open_file(1);
 #endif
 	display_main_list();
 	return i;
-    } /* while (1) */
+    } /* while (TRUE) */
 }
 
 int do_writeout_void(void)
 {
-    return do_writeout(filename, 0, 0);
+    return do_writeout(FALSE);
 }
 
 /* Return a malloc()ed string containing the actual directory, used
@@ -2249,7 +2345,7 @@ char *input_tab(char *buf, int place, int *lastwastab, int *newplace, int *list)
 		 pos <= strlen(matches[0]); pos++)
 		tmp++;
 
-	    while (1) {
+	    while (TRUE) {
 		match_matches = 0;
 
 		for (i = 0; i < num_matches; i++) {
@@ -2339,16 +2435,24 @@ char *input_tab(char *buf, int place, int *lastwastab, int *newplace, int *list)
 }
 #endif /* !DISABLE_TABCOMP */
 
-#ifndef DISABLE_BROWSER
-/* Return the stat of the file pointed to by path */
-struct stat filestat(const char *path)
+#if !defined(DISABLE_BROWSER) || !defined(NANO_SMALL)
+/* Only print the last part of a path; isn't there a shell
+ * command for this? */
+const char *tail(const char *foo)
 {
-    struct stat st;
+    const char *tmp = foo + strlen(foo);
 
-    stat(path, &st);
-    return st;
+    while (*tmp != '/' && tmp != foo)
+	tmp--;
+
+    if (*tmp == '/')
+	tmp++;
+
+    return tmp;
 }
+#endif
 
+#ifndef DISABLE_BROWSER
 /* Our sort routine for file listings -- sort directories before
  * files, and then alphabetically. */ 
 int diralphasort(const void *va, const void *vb)
@@ -2374,21 +2478,6 @@ void free_charptrarray(char **array, int len)
     free(array);
 }
 
-/* Only print the last part of a path; isn't there a shell 
- * command for this? */
-const char *tail(const char *foo)
-{
-    const char *tmp = foo + strlen(foo);
-
-    while (*tmp != '/' && tmp != foo)
-	tmp--;
-
-    if (*tmp == '/')
-	tmp++;
-
-    return tmp;
-}
-
 /* Strip one dir from the end of a string. */
 void striponedir(char *foo)
 {
@@ -2478,7 +2567,7 @@ char *do_browser(const char *inpath)
     struct stat st;
     char *foo, *retval = NULL;
     static char *path = NULL;
-    int numents = 0, i = 0, j = 0, kbinput = -1, meta, longest = 0;
+    int numents = 0, i = 0, j = 0, kbinput = ERR, meta_key, longest = 0;
     int abort = 0, col = 0, selected = 0, editline = 0, width = 0;
     int filecols = 0, lineno = 0;
     char **filelist = (char **)NULL;
@@ -2568,7 +2657,7 @@ char *do_browser(const char *inpath)
 
             break;
 #endif
-	case NANO_UP_KEY:
+	case NANO_PREVLINE_KEY:
 	    if (selected - width >= 0)
 		selected -= width;
 	    break;
@@ -2576,7 +2665,7 @@ char *do_browser(const char *inpath)
 	    if (selected > 0)
 		selected--;
 	    break;
-	case NANO_DOWN_KEY:
+	case NANO_NEXTLINE_KEY:
 	    if (selected + width <= numents - 1)
 		selected += width;
 	    break;
@@ -2670,9 +2759,9 @@ char *do_browser(const char *inpath)
 	case 'G': /* Pico compatibility */
 	case 'g':
 	    curs_set(1);
-	    j = statusq(0, gotodir_list, "",
+	    j = statusq(FALSE, gotodir_list, "",
 #ifndef NANO_SMALL
-		0,
+		NULL,
 #endif
 		_("Goto Directory"));
 	    bottombars(browser_list);
@@ -2751,7 +2840,7 @@ char *do_browser(const char *inpath)
 		if (S_ISLNK(st.st_mode)) {
 		     /* Aha!  It's a symlink!  Now, is it a dir?  If so,
 			mark it as such */
-		    st = filestat(filelist[j]);
+		    stat(filelist[j], &st);
 		    if (S_ISDIR(st.st_mode))
 			strcpy(foo + longest - 5, "(dir)");
 		    else
@@ -2792,7 +2881,7 @@ char *do_browser(const char *inpath)
 	    }
 	}
  	wrefresh(edit);
-    } while ((kbinput = get_kbinput(edit, &meta)) != NANO_EXIT_KEY && kbinput != NANO_EXIT_FKEY);
+    } while ((kbinput = get_kbinput(edit, &meta_key)) != NANO_EXIT_KEY && kbinput != NANO_EXIT_FKEY);
     curs_set(1);
     blank_edit();
     titlebar(NULL);
