@@ -1,4 +1,4 @@
-/* $Id: global.c,v 1.7 2004/01/09 23:04:55 dolorous Exp $ */
+/* $Id: global.c,v 1.19 2004/03/19 21:46:34 dolorous Exp $ */
 /**************************************************************************
  *   global.c                                                             *
  *                                                                        *
@@ -59,7 +59,6 @@ filestruct *fileage = NULL;	/* Our file buffer */
 filestruct *edittop = NULL;	/* Pointer to the top of the edit
 				   buffer with respect to the
 				   file struct */
-filestruct *editbot = NULL;	/* Same for the bottom */
 filestruct *filebot = NULL;	/* Last node in the file struct */
 filestruct *cutbuffer = NULL;	/* A place to store cut text */
 
@@ -72,6 +71,10 @@ char *quotestr = NULL;		/* Quote string.  The default value is
 				   set in main(). */
 #endif
 
+#ifndef NANO_SMALL
+char *backup_dir = NULL;	/* Backup directory. */
+#endif
+
 int resetstatuspos;		/* Hack for resetting the status bar 
 				   cursor position */
 char *answer = NULL;		/* Answer str to many questions */
@@ -158,7 +161,7 @@ int curses_ended = FALSE;	/* Indicates to statusbar() to simply
 				 * ended curses mode. */
 
 
-int length_of_list(const shortcut *s) 
+int length_of_list(const shortcut *s)
 {
     int i = 0;
 
@@ -168,11 +171,12 @@ int length_of_list(const shortcut *s)
 }
 
 /* Initialize a struct *without* our lovely braces =( */
-void sc_init_one(shortcut **shortcutage, int key, const char *desc,
+void sc_init_one(shortcut **shortcutage, int ctrlval, const char *desc,
 #ifndef DISABLE_HELP
 	const char *help,
 #endif
-	int meta, int func_key, int misc, int view, int (*func) (void))
+	int metaval, int funcval, int miscval, int view, int
+	(*func)(void))
 {
     shortcut *s;
 
@@ -186,14 +190,14 @@ void sc_init_one(shortcut **shortcutage, int key, const char *desc,
 	s = s->next; 
     }
 
-    s->val = key;
+    s->ctrlval = ctrlval;
     s->desc = desc;
 #ifndef DISABLE_HELP
     s->help = help;
 #endif
-    s->metaval = meta;
-    s->func_key = func_key;
-    s->misc = misc;
+    s->metaval = metaval;
+    s->funcval = funcval;
+    s->miscval = miscval;
     s->viewok = view;
     s->func = func;
     s->next = NULL;
@@ -328,8 +332,8 @@ void shortcut_init(int unjustify)
 	"", *nano_whereis_next_msg = "", *nano_prevpage_msg =
 	"", *nano_nextpage_msg = "", *nano_cut_msg =
 	"", *nano_uncut_msg = "", *nano_cursorpos_msg =
-	"", *nano_spell_msg = "", *nano_up_msg =
-	"", *nano_down_msg = "", *nano_forward_msg =
+	"", *nano_spell_msg = "", *nano_prevline_msg =
+	"", *nano_nextline_msg = "", *nano_forward_msg =
 	"", *nano_back_msg = "", *nano_home_msg =
 	"", *nano_end_msg = "", *nano_firstline_msg =
 	"", *nano_lastline_msg = "", *nano_refresh_msg =
@@ -374,8 +378,8 @@ void shortcut_init(int unjustify)
     nano_uncut_msg = _("Uncut from the cutbuffer into the current line");
     nano_cursorpos_msg = _("Show the position of the cursor");
     nano_spell_msg = _("Invoke the spell checker, if available");
-    nano_up_msg = _("Move up one line");
-    nano_down_msg = _("Move down one line");
+    nano_prevline_msg = _("Move to the previous line");
+    nano_nextline_msg = _("Move to the next line");
     nano_forward_msg = _("Move forward one character");
     nano_back_msg = _("Move back one character");
     nano_home_msg = _("Move to the beginning of the current line");
@@ -424,10 +428,10 @@ void shortcut_init(int unjustify)
 /* The following macro is to be used in calling sc_init_one.  The point is
  * that sc_init_one takes 9 arguments, unless DISABLE_HELP is defined,
  * when the fourth one should not be there. */
-#ifdef DISABLE_HELP
-#  define IFHELP(help, nextvar) nextvar
-#else
+#ifndef DISABLE_HELP
 #  define IFHELP(help, nextvar) help, nextvar
+#else
+#  define IFHELP(help, nextvar) nextvar
 #endif
 
     /* Translators: try to keep this string under 10 characters long */
@@ -494,8 +498,8 @@ void shortcut_init(int unjustify)
     if (unjustify)
     /* Translators: try to keep this string under 10 characters long */
 	sc_init_one(&main_list, NANO_UNJUSTIFY_KEY, _("UnJustify"),
-		IFHELP(nano_unjustify_msg, NANO_NO_KEY), NANO_NO_KEY,
-		NANO_NO_KEY, NOVIEW, do_uncut_text);
+		IFHELP(nano_unjustify_msg, NANO_NO_KEY), NANO_UNJUSTIFY_FKEY,
+		NANO_NO_KEY, NOVIEW, 0);
     else
     /* Translators: try to keep this string under 10 characters long */
 	sc_init_one(&main_list, NANO_UNCUT_KEY, _("UnCut Txt"),
@@ -520,12 +524,12 @@ void shortcut_init(int unjustify)
 		IFHELP(nano_replace_msg, NANO_ALT_REPLACE_KEY), NANO_REPLACE_FKEY,
 		NANO_NO_KEY, NOVIEW, do_replace);
 
-    sc_init_one(&main_list, NANO_UP_KEY, _("Up"),
-		IFHELP(nano_up_msg, NANO_NO_KEY), NANO_NO_KEY,
+    sc_init_one(&main_list, NANO_PREVLINE_KEY, _("Prev Line"),
+		IFHELP(nano_prevline_msg, NANO_NO_KEY), NANO_NO_KEY,
 		NANO_NO_KEY, VIEW, do_up);
 
-    sc_init_one(&main_list, NANO_DOWN_KEY, _("Down"),
-		IFHELP(nano_down_msg, NANO_NO_KEY), NANO_NO_KEY,
+    sc_init_one(&main_list, NANO_NEXTLINE_KEY, _("Next Line"),
+		IFHELP(nano_nextline_msg, NANO_NO_KEY), NANO_NO_KEY,
 		NANO_NO_KEY, VIEW, do_down);
 
     sc_init_one(&main_list, NANO_FORWARD_KEY, _("Forward"),
@@ -580,7 +584,7 @@ void shortcut_init(int unjustify)
 
     sc_init_one(&main_list, NANO_NO_KEY, _("Verbatim Input"),
 		IFHELP(nano_verbatim_msg, NANO_VERBATIM_KEY), NANO_NO_KEY,
-		NANO_NO_KEY, VIEW, do_verbatim_input);
+		NANO_NO_KEY, NOVIEW, do_verbatim_input);
 
 #ifdef ENABLE_MULTIBUFFER
     sc_init_one(&main_list, NANO_NO_KEY, _("Previous File"),
@@ -615,12 +619,12 @@ void shortcut_init(int unjustify)
 
     /* Translators: try to keep this string under 10 characters long */
     sc_init_one(&whereis_list, NANO_FIRSTLINE_KEY, _("First Line"),
-		IFHELP(nano_firstline_msg, NANO_NO_KEY), NANO_NO_KEY,
+		IFHELP(nano_firstline_msg, NANO_NO_KEY), NANO_FIRSTLINE_FKEY,
 		NANO_NO_KEY, VIEW, do_first_line);
 
     /* Translators: try to keep this string under 10 characters long */
     sc_init_one(&whereis_list, NANO_LASTLINE_KEY, _("Last Line"),
-		IFHELP(nano_lastline_msg, NANO_NO_KEY), NANO_NO_KEY,
+		IFHELP(nano_lastline_msg, NANO_NO_KEY), NANO_LASTLINE_FKEY,
 		NANO_NO_KEY, VIEW, do_last_line);
 
     /* Translators: try to keep this string under 10 characters long */
@@ -716,7 +720,7 @@ void shortcut_init(int unjustify)
 
     sc_init_one(&replace_list, NANO_HISTORY_KEY, _("History"),
 		IFHELP(nano_editstr_msg, NANO_NO_KEY), NANO_NO_KEY,
-		NANO_DOWN_KEY, VIEW, 0);
+		NANO_NO_KEY, VIEW, 0);
 #endif /* !NANO_SMALL */
 
     free_shortcutage(&replace_list_2);
@@ -738,9 +742,9 @@ void shortcut_init(int unjustify)
 		NANO_NO_KEY, VIEW, do_last_line);
 
 #ifndef NANO_SMALL
-    sc_init_one(&replace_list_2, NANO_UP_KEY, _("History"),
+    sc_init_one(&replace_list_2, NANO_HISTORY_KEY, _("History"),
 		IFHELP(nano_editstr_msg, NANO_NO_KEY), NANO_NO_KEY,
-		NANO_DOWN_KEY, VIEW, 0);
+		NANO_NO_KEY, VIEW, 0);
 #endif
 
     free_shortcutage(&goto_list);
@@ -932,6 +936,10 @@ void thanks_for_all_the_fish(void)
     if (quotestr != NULL)
 	free(quotestr);
 #endif
+#ifndef NANO_SMALL
+    if (backup_dir != NULL)
+        free(backup_dir);
+#endif
 #ifndef DISABLE_OPERATINGDIR
     if (operating_dir != NULL)
 	free(operating_dir);
