@@ -1,4 +1,4 @@
-/* $Id: nano.h,v 1.15 2004/01/09 23:04:55 dolorous Exp $ */
+/* $Id: nano.h,v 1.26 2004/03/30 04:17:10 dolorous Exp $ */
 /**************************************************************************
  *   nano.h                                                               *
  *                                                                        *
@@ -19,13 +19,21 @@
  *                                                                        *
  **************************************************************************/
 
+#ifndef NANO_H
+#define NANO_H 1
+
+#ifdef __TANDEM
+/* Tandem NonStop Kernel */
+#include <floss.h>
+#define NANO_ROOT_UID 65535
+#else
+#define NANO_ROOT_UID 0
+#endif
+
 #ifdef HAVE_LIMITS_H
 #include <limits.h>
 #endif
 
-#ifndef NANO_H
-#define NANO_H 1
-
 /* Macros for the flags int... */
 #define SET(bit) flags |= bit
 #define UNSET(bit) flags &= ~bit
@@ -179,11 +187,11 @@ typedef struct openfilestruct {
 
 typedef struct shortcut {
     /* Key values that aren't used should be set to NANO_NO_KEY. */
-    int val;		/* Special sentinel key or control key we want
+    int ctrlval;	/* Special sentinel key or control key we want
 			 * bound. */
-    int metaval;		/* Meta key we want bound. */
-    int func_key;	/* Function key we want bound. */
-    int misc;		/* Other Meta key we want bound. */
+    int metaval;	/* Meta key we want bound. */
+    int funcval;	/* Function key we want bound. */
+    int miscval;	/* Other Meta key we want bound. */
     int viewok;		/* Is this function legal in view mode? */
     int (*func) (void);	/* Function to call when we catch this key. */
     const char *desc;	/* Description, e.g. "Page Up". */
@@ -288,7 +296,6 @@ typedef struct historyheadtype {
 #define PRESERVE		(1<<27)
 #define HISTORY_CHANGED		(1<<28)
 #define HISTORYLOG		(1<<29)
-#define JUSTIFY_MODE		(1<<30)
 
 /* Control key sequences, changing these would be very very bad. */
 #define NANO_CONTROL_SPACE 0
@@ -399,14 +406,17 @@ typedef struct historyheadtype {
 #define NANO_SPELL_KEY		NANO_CONTROL_T
 #define NANO_SPELL_FKEY		KEY_F(12)
 #define NANO_FIRSTLINE_KEY	NANO_PREVPAGE_KEY
+#define NANO_FIRSTLINE_FKEY	NANO_PREVPAGE_FKEY
 #define NANO_LASTLINE_KEY	NANO_NEXTPAGE_KEY
+#define NANO_LASTLINE_FKEY	NANO_NEXTPAGE_FKEY
 #define NANO_CANCEL_KEY		NANO_CONTROL_C
 #define NANO_REFRESH_KEY	NANO_CONTROL_L
 #define NANO_JUSTIFY_KEY	NANO_CONTROL_J
 #define NANO_JUSTIFY_FKEY	KEY_F(4)
-#define NANO_UNJUSTIFY_KEY	NANO_CONTROL_U
-#define NANO_UP_KEY		NANO_CONTROL_P
-#define NANO_DOWN_KEY		NANO_CONTROL_N
+#define NANO_UNJUSTIFY_KEY	NANO_UNCUT_KEY	/* Same key as uncut. */
+#define NANO_UNJUSTIFY_FKEY	NANO_UNCUT_FKEY	/* Same key as uncut. */
+#define NANO_PREVLINE_KEY	NANO_CONTROL_P
+#define NANO_NEXTLINE_KEY	NANO_CONTROL_N
 #define NANO_FORWARD_KEY	NANO_CONTROL_F
 #define NANO_BACK_KEY		NANO_CONTROL_B
 #define NANO_MARK_KEY		NANO_CONTROL_6
