@@ -1,4 +1,4 @@
-/* $Id: proto.h,v 1.14 2004/01/09 23:04:55 dolorous Exp $ */
+/* $Id: proto.h,v 1.26 2004/03/30 04:17:10 dolorous Exp $ */
 /**************************************************************************
  *   proto.h                                                              *
  *                                                                        *
@@ -50,6 +50,10 @@ extern int currslen;
 extern char *quotestr;
 #endif
 
+#ifndef NANO_SMALL
+extern char *backup_dir;
+#endif
+
 extern WINDOW *edit, *topwin, *bottomwin;
 extern char *filename;
 extern struct stat originalfilestat;
@@ -70,7 +74,7 @@ extern char *alt_speller;
 
 extern int resetstatuspos;
 extern struct stat fileinfo;
-extern filestruct *current, *fileage, *edittop, *editbot, *filebot; 
+extern filestruct *current, *fileage, *edittop, *filebot;
 extern filestruct *cutbuffer;
 #ifndef NANO_SMALL
 extern filestruct *mark_beginbuf;
@@ -139,7 +143,7 @@ void update_color(void);
 
 /* Public functions in cut.c */
 filestruct *get_cutbottom(void);
-void add_to_cutbuffer(filestruct *inptr);
+void add_to_cutbuffer(filestruct *inptr, int allow_concat);
 void cut_marked_segment(void);
 int do_cut_text(void);
 int do_uncut_text(void);
@@ -178,8 +182,15 @@ char *safe_tempnam(const char *dirname, const char *filename_prefix);
 void init_operating_dir(void);
 int check_operating_dir(const char *currpath, int allow_tabcomp);
 #endif
+#ifndef NANO_SMALL
+void init_backup_dir(void);
+#endif
 int write_file(const char *name, int tmp, int append, int nonamechange);
-int do_writeout(const char *path, int exiting, int append);
+#ifndef NANO_SMALL
+int write_marked(const char *name, int tmp, int append, int
+	nonamechange);
+#endif
+int do_writeout(int exiting);
 int do_writeout_void(void);
 char *real_dir_from_tilde(const char *buf);
 #ifndef DISABLE_TABCOMP
@@ -188,11 +199,12 @@ char **username_tab_completion(char *buf, int *num_matches);
 char **cwd_tab_completion(char *buf, int *num_matches);
 char *input_tab(char *buf, int place, int *lastwastab, int *newplace, int *list);
 #endif
+#if !defined(DISABLE_BROWSER) || !defined(NANO_SMALL)
+const char *tail(const char *foo);
+#endif
 #ifndef DISABLE_BROWSER
-struct stat filestat(const char *path);
 int diralphasort(const void *va, const void *vb);
 void free_charptrarray(char **array, int len);
-const char *tail(const char *foo);
 void striponedir(char *foo);
 int readable_dir(const char *path);
 char **browser_init(const char *path, int *longest, int *numents);
@@ -206,7 +218,8 @@ void sc_init_one(shortcut **shortcutage, int key, const char *desc,
 #ifndef DISABLE_HELP
 	const char *help,
 #endif
-	int meta, int func_key, int misc, int view, int (*func) (void));
+	int metaval, int funcval, int miscval, int view, int
+	(*func)(void));
 #ifndef NANO_SMALL
 void toggle_init_one(int val, const char *desc, int flag);
 void toggle_init(void);
@@ -313,13 +326,12 @@ filestruct *backup_lines(filestruct *first_line, size_t par_len,
 			size_t quote_len);
 int breakable(const char *line, int goal);
 int break_line(const char *line, int goal, int force);
-int do_para_operation(int operation);
-#endif /* !DISABLE_JUSTIFY */
-int do_justify(void);
-#ifndef DISABLE_JUSTIFY
+int do_para_search(int search_type, size_t *quote, size_t *par, size_t
+	*indent, int do_refresh);
 int do_para_begin(void);
 int do_para_end(void);
-#endif
+#endif /* !DISABLE_JUSTIFY */
+int do_justify(void);
 int do_exit(void);
 void signal_init(void);
 RETSIGTYPE handle_hupterm(int signal);
@@ -327,6 +339,7 @@ RETSIGTYPE do_suspend(int signal);
 RETSIGTYPE do_cont(int signal);
 #ifndef NANO_SMALL
 void handle_sigwinch(int s);
+void allow_pending_sigwinch(int allow);
 #endif
 void print_numlock_warning(void);
 #ifndef NANO_SMALL
@@ -359,19 +372,19 @@ void not_found_msg(const char *str);
 void search_abort(void);
 void search_init_globals(void);
 int search_init(int replacing);
-int is_whole_word(int curr_pos, const char *datastr, const char *searchword);
-filestruct *findnextstr(int quiet, int bracket_mode,
-			const filestruct *begin, int beginx,
-			const char *needle, int no_sameline);
+int is_whole_word(int curr_pos, const char *datastr, const char
+	*searchword);
+int findnextstr(int can_display_wrap, int wholeword, const filestruct
+	*begin, size_t beginx, const char *needle, int no_sameline);
 int do_search(void);
 int do_research(void);
 void replace_abort(void);
 #ifdef HAVE_REGEX_H
 int replace_regexp(char *string, int create_flag);
 #endif
-char *replace_line(void);
-int do_replace_loop(const char *prevanswer, const filestruct *begin,
-			int *beginx, int wholewords, int *i);
+char *replace_line(const char *needle);
+int do_replace_loop(const char *needle, const filestruct *real_current,
+	size_t *real_current_x, int wholewords);
 int do_replace(void);
 int do_gotoline(int line, int save_pos);
 int do_gotoline_void(void);
@@ -396,10 +409,13 @@ void save_history(void);
 #endif
 
 /* Public functions in utils.c */
+#ifdef HAVE_REGEX_H
 #ifdef BROKEN_REGEXEC
 int regexec_safe(const regex_t *preg, const char *string, size_t nmatch,
 	regmatch_t pmatch[], int eflags);
 #endif
+int regexp_bol_or_eol(const regex_t *preg, const char *string);
+#endif
 int is_cntrl_char(int c);
 int num_of_digits(int n);
 void align(char **strp);
@@ -420,7 +436,7 @@ const char *revstristr(const char *haystack, const char *needle,
 #endif
 const char *stristr(const char *haystack, const char *needle);
 const char *strstrwrapper(const char *haystack, const char *needle,
-			const char *rev_start, int line_pos);
+	const char *start);
 void nperror(const char *s);
 void *nmalloc(size_t howmuch);
 void *nrealloc(void *ptr, size_t howmuch);
@@ -435,14 +451,15 @@ int check_wildcard_match(const char *text, const char *pattern);
 #endif
 
 /* Public functions in winio.c */
-int get_kbinput(WINDOW *win, int *meta);
-char *get_verbatim_kbinput(WINDOW *win, int *kbinput_len,
-	int allow_ascii);
+int get_kbinput(WINDOW *win, int *meta_key);
+int *get_verbatim_kbinput(WINDOW *win, int *kbinput_len, int
+	allow_ascii);
 int get_ignored_kbinput(WINDOW *win);
-int get_accepted_kbinput(WINDOW *win, int kbinput, int *meta);
+int get_accepted_kbinput(WINDOW *win, int kbinput, int *meta_key);
 int get_ascii_kbinput(WINDOW *win, int kbinput);
-int get_escape_seq_kbinput(WINDOW *win, char *escape_seq, int
+int get_escape_seq_kbinput(WINDOW *win, int *escape_seq, int
 	escape_seq_len);
+int get_escape_seq_abcd(int kbinput);
 int get_mouseinput(int *mouse_x, int *mouse_y, int shortcut);
 int do_first_line(void);
 int do_last_line(void);
@@ -487,7 +504,7 @@ void center_cursor(void);
 void edit_refresh(void);
 void edit_refresh_clearok(void);
 void edit_update(filestruct *fileptr, topmidnone location);
-int statusq(int tabs, const shortcut *s, const char *def,
+int statusq(int allowtabs, const shortcut *s, const char *def,
 #ifndef NANO_SMALL
 		historyheadtype *history_list,
 #endif
@@ -501,7 +518,6 @@ int do_cursorpos_void(void);
 int line_len(const char *ptr);
 int do_help(void);
 void do_replace_highlight(int highlight_flag, const char *word);
-void fix_editbot(void);
 #ifdef DEBUG
 void dump_buffer(const filestruct *inptr);
 void dump_buffer_reverse(void);
