@@ -1,4 +1,4 @@
-/* $Id: winio.c,v 1.22 2004/01/09 23:06:54 dolorous Exp $ */
+/* $Id: winio.c,v 1.51 2004/03/29 23:09:08 dolorous Exp $ */
 /**************************************************************************
  *   winio.c                                                              *
  *                                                                        *
@@ -30,33 +30,61 @@
 #include "proto.h"
 #include "nano.h"
 
+#ifdef NANO_EXTRA
+#include <time.h>
+#endif
+
 static int statblank = 0;	/* Number of keystrokes left after
 				   we call statusbar(), before we
 				   actually blank the statusbar */
 
 /* Read in a single input character.  If it's ignored, swallow it and go
- * on.  Otherwise, try to translate it from ASCII and extended (keypad)
- * input.  Assume nodelay(win) is FALSE. */
-int get_kbinput(WINDOW *win, int *meta)
+ * on.  Otherwise, try to translate it from ASCII, extended keypad
+ * values, and/or escape sequences.  Supported extended keypad values
+ * consist of [arrow key], Ctrl-[arrow key], Shift-[arrow key], Enter,
+ * Backspace, Insert, Delete, Home, End, PageUp, PageDown, and F1-F14.
+ * Assume nodelay(win) is FALSE. */
+int get_kbinput(WINDOW *win, int *meta_key)
 {
     int kbinput, retval;
 
+#ifndef NANO_SMALL
+    allow_pending_sigwinch(TRUE);
+#endif
+
     kbinput = get_ignored_kbinput(win);
-    retval = get_accepted_kbinput(win, kbinput, meta);
+    retval = get_accepted_kbinput(win, kbinput, meta_key);
+
+#ifndef NANO_SMALL
+    allow_pending_sigwinch(FALSE);
+#endif
 
     return retval;
 }
 
-/* Read in a string of input characters (e. g. an escape sequence)
+/* Read in a string of input characters (e.g. an escape sequence)
  * verbatim, and return the length of the string in kbinput_len.  Assume
  * nodelay(win) is FALSE. */
-char *get_verbatim_kbinput(WINDOW *win, int *kbinput_len,
-	int allow_ascii)
+int *get_verbatim_kbinput(WINDOW *win, int *kbinput_len, int
+	allow_ascii)
 {
-    char *verbatim_kbinput;
-    int kbinput = wgetch(win);
+    int kbinput, *verbatim_kbinput;
 
-    verbatim_kbinput = charalloc(1);
+#ifndef NANO_SMALL
+    allow_pending_sigwinch(TRUE);
+#endif
+
+    /* Switch to raw mode so that we can type ^C, ^Q, ^S, ^Z, and ^\
+     * (and ^Y on the Hurd) without getting interrupts, and Turn the
+     * keypad off so that we don't get extended keypad values all of
+     * which are outside the ASCII range. */
+#ifdef _POSIX_VDISABLE
+    raw();
+#endif
+    keypad(win, FALSE);
+
+    kbinput = wgetch(win);
+    verbatim_kbinput = (int *)nmalloc(sizeof(int));
     verbatim_kbinput[0] = kbinput;
     *kbinput_len = 1;
 
@@ -64,20 +92,32 @@ char *get_verbatim_kbinput(WINDOW *win, int *kbinput_len,
 	/* Entering a three-digit decimal ASCII code from 000-255 in
 	 * verbatim mode will produce the corresponding ASCII
 	 * character. */
-	verbatim_kbinput[0] = (char)get_ascii_kbinput(win, kbinput);
+	verbatim_kbinput[0] = get_ascii_kbinput(win, kbinput);
     else {
 	nodelay(win, TRUE);
 	while ((kbinput = wgetch(win)) != ERR) {
 	    (*kbinput_len)++;
-	    verbatim_kbinput = charealloc(verbatim_kbinput, *kbinput_len);
-	    verbatim_kbinput[*kbinput_len - 1] = (char)kbinput;
+	    verbatim_kbinput = realloc(verbatim_kbinput, *kbinput_len * sizeof(int));
+	    verbatim_kbinput[*kbinput_len - 1] = kbinput;
 	}
 	nodelay(win, FALSE);
     }
 
+    /* Switch back to cbreak mode and turn the keypad back on now that
+     * we're done. */
+#ifdef _POSIX_VDISABLE
+    cbreak();
+#endif
+    keypad(win, TRUE);
+
 #ifdef DEBUG
     fprintf(stderr, "get_verbatim_kbinput(): verbatim_kbinput = %s\n", verbatim_kbinput);
 #endif
+
+#ifndef NANO_SMALL
+    allow_pending_sigwinch(FALSE);
+#endif
+
     return verbatim_kbinput;
 }
 
@@ -87,7 +127,7 @@ int get_ignored_kbinput(WINDOW *win)
 {
     int kbinput;
 
-    while (1) {
+    while (TRUE) {
 	kbinput = wgetch(win);
 	switch (kbinput) {
 	    case ERR:
@@ -110,20 +150,22 @@ int get_ignored_kbinput(WINDOW *win)
     }
 }
 
-/* Translate acceptable ASCII and extended (keypad) input.  Set meta to
- * 1 if we get a Meta sequence.  Assume nodelay(win) is FALSE. */
-int get_accepted_kbinput(WINDOW *win, int kbinput, int *meta)
+/* Translate acceptable ASCII, extended keypad values, and/or escape
+ * sequences.  Set meta_key to 1 if we get a Meta sequence.  Assume
+ * nodelay(win) is FALSE. */
+int get_accepted_kbinput(WINDOW *win, int kbinput, int *meta_key)
 {
-    *meta = 0;
+    *meta_key = FALSE;
 
     switch (kbinput) {
 	case NANO_CONTROL_3: /* Escape */
-	    switch (kbinput = wgetch(win)) {
+	    kbinput = wgetch(win);
+	    switch (kbinput) {
 		case NANO_CONTROL_3: /* Escape */
 		    kbinput = wgetch(win);
 		    /* Esc Esc [three-digit decimal ASCII code from
 		     * 000-255] == [corresponding ASCII character];
-		       Esc Esc 2 obviously can't be Ctrl-2 here */
+		     * Esc Esc 2 obviously can't be Ctrl-2 here */
 		    if (kbinput >= '0' && kbinput <= '2')
 			kbinput = get_ascii_kbinput(win, kbinput);
 		    /* Esc Esc [character] == Ctrl-[character];
@@ -151,14 +193,13 @@ int get_accepted_kbinput(WINDOW *win, int kbinput, int *meta)
 		 * Hemel. */
 		case '[':
 		{
-		    int old_kbinput = kbinput, escape_seq_len;
-		    char *escape_seq;
+		    int old_kbinput = kbinput, *escape_seq, escape_seq_len;
 		    nodelay(win, TRUE);
 		    kbinput = wgetch(win);
 		    switch (kbinput) {
 			case ERR:
 			    kbinput = tolower(old_kbinput);
-			    *meta = 1;
+			    *meta_key = TRUE;
 			    break;
 			default:
 			    ungetch(kbinput);
@@ -173,7 +214,7 @@ int get_accepted_kbinput(WINDOW *win, int kbinput, int *meta)
 		default:
 		    /* Esc [character] == Meta-[character] */
 		    kbinput = tolower(kbinput);
-		    *meta = 1;
+		    *meta_key = TRUE;
 	    }
 	    break;
 	case NANO_CONTROL_8:
@@ -182,10 +223,10 @@ int get_accepted_kbinput(WINDOW *win, int kbinput, int *meta)
 	    kbinput = ISSET(REBIND_DELETE) ? NANO_DELETE_KEY : NANO_BACKSPACE_KEY;
 	    break;
 	case KEY_DOWN:
-	    kbinput = NANO_DOWN_KEY;
+	    kbinput = NANO_NEXTLINE_KEY;
 	    break;
 	case KEY_UP:
-	    kbinput = NANO_UP_KEY;
+	    kbinput = NANO_PREVLINE_KEY;
 	    break;
 	case KEY_LEFT:
 	    kbinput = NANO_BACK_KEY;
@@ -231,7 +272,7 @@ int get_accepted_kbinput(WINDOW *win, int kbinput, int *meta)
 	    break;
     }
 #ifdef DEBUG
-    fprintf(stderr, "get_accepted_kbinput(): kbinput = %d, meta = %d\n", kbinput, *meta);
+    fprintf(stderr, "get_accepted_kbinput(): kbinput = %d, meta_key = %d\n", kbinput, *meta_key);
 #endif
     return kbinput;
 }
@@ -302,29 +343,59 @@ int get_ascii_kbinput(WINDOW *win, int kbinput)
     return retval;
 }
 
-/* Translate common escape sequences for some keys.  These are generated
- * when the terminal doesn't support those keys.  Assume that Escape has
- * already been read in, and that nodelay(win) is TRUE. */
-int get_escape_seq_kbinput(WINDOW *win, char *escape_seq, int
+/* Translate escape sequences, most of which correspond to extended
+ * keypad values.  These sequences are generated when the terminal
+ * doesn't support the needed keys.  Assume that Escape has already been
+ * read in, and that nodelay(win) is TRUE.
+ *
+ * The supported terminals are the Linux console, the FreeBSD console,
+ * the Hurd console (a.k.a. the Mach console), xterm, rxvt, and Eterm.
+ * There are several escape sequence conflicts and omissions, outlined
+ * as follows:
+ *
+ * - F1 on FreeBSD console == kmous on xterm/rxvt/Eterm; the latter is
+ *   omitted.  (Mouse input will only work properly if the extended
+ *   keypad value KEY_MOUSE is generated on mouse events instead of the
+ *   kmous escape sequence.)
+ * - F9 on FreeBSD console == PageDown on Hurd console; the former is
+ *   omitted.  (The editing keypad, consisting of Insert, Delete, Home,
+ *   End, PageUp, and PageDown, is more important to have working than
+ *   the function keys, because the functions of the former are not
+ *   arbitrary and the functions of the latter are.)
+ * - F10 on FreeBSD console == PageUp on Hurd console; the former is
+ *   omitted.  (Same as above.)
+ * - F13 on FreeBSD console == End on Hurd console; the former is
+ *   omitted.  (Same as above.)
+ * - The Hurd console has no escape sequences for F11, F12, F13, or
+ *   F14. */
+int get_escape_seq_kbinput(WINDOW *win, int *escape_seq, int
 	escape_seq_len)
 {
-    int kbinput = -1;
+    int kbinput = ERR;
 
     if (escape_seq_len > 1) {
 	switch (escape_seq[0]) {
 	    case 'O':
 		switch (escape_seq[1]) {
+		    case '2':
+			if (escape_seq_len >= 3) {
+			    switch (escape_seq[2]) {
+				case 'P': /* Esc O 2 P == F13 on
+					   * xterm. */
+				    kbinput = KEY_F(13);
+				    break;
+				case 'Q': /* Esc O 2 Q == F14 on
+					   * xterm. */
+				    kbinput = KEY_F(14);
+				    break;
+			    }
+			}
+			break;
 		    case 'A': /* Esc O A == Up on xterm. */
-			kbinput = NANO_UP_KEY;
-			break;
 		    case 'B': /* Esc O B == Down on xterm. */
-			kbinput = NANO_DOWN_KEY;
-			break;
 		    case 'C': /* Esc O C == Right on xterm. */
-			kbinput = NANO_FORWARD_KEY;
-			break;
 		    case 'D': /* Esc O D == Left on xterm. */
-			kbinput = NANO_BACK_KEY;
+			kbinput = get_escape_seq_abcd(escape_seq[1]);
 			break;
 		    case 'F': /* Esc O F == End on xterm. */
 			kbinput = NANO_END_KEY;
@@ -332,96 +403,169 @@ int get_escape_seq_kbinput(WINDOW *win, char *escape_seq, int
 		    case 'H': /* Esc O H == Home on xterm. */
 			kbinput = NANO_HOME_KEY;
 			break;
+		    case 'P': /* Esc O P == F1 on Hurd console. */
+			kbinput = KEY_F(1);
+			break;
+		    case 'Q': /* Esc O Q == F2 on Hurd console. */
+			kbinput = KEY_F(2);
+			break;
+		    case 'R': /* Esc O R == F3 on Hurd console. */
+			kbinput = KEY_F(3);
+			break;
+		    case 'S': /* Esc O S == F4 on Hurd console. */
+			kbinput = KEY_F(4);
+			break;
+		    case 'T': /* Esc O T == F5 on Hurd console. */
+			kbinput = KEY_F(5);
+			break;
+		    case 'U': /* Esc O U == F6 on Hurd console. */
+			kbinput = KEY_F(6);
+			break;
+		    case 'V': /* Esc O V == F7 on Hurd console. */
+			kbinput = KEY_F(7);
+			break;
+		    case 'W': /* Esc O W == F8 on Hurd console. */
+			kbinput = KEY_F(8);
+			break;
+		    case 'X': /* Esc O X == F9 on Hurd console. */
+			kbinput = KEY_F(9);
+			break;
+		    case 'Y': /* Esc O Y == F10 on Hurd console. */
+			kbinput = KEY_F(10);
+			break;
 		    case 'a': /* Esc O a == Ctrl-Up on rxvt. */
-			kbinput = NANO_UP_KEY;
-			break;
 		    case 'b': /* Esc O b == Ctrl-Down on rxvt. */
-			kbinput = NANO_DOWN_KEY;
-			break;
 		    case 'c': /* Esc O c == Ctrl-Right on rxvt. */
-			kbinput = NANO_FORWARD_KEY;
-			break;
 		    case 'd': /* Esc O d == Ctrl-Left on rxvt. */
-			kbinput = NANO_BACK_KEY;
+			kbinput = get_escape_seq_abcd(escape_seq[1]);
 			break;
 		}
 		break;
 	    case 'o':
 		switch (escape_seq[1]) {
 		    case 'a': /* Esc o a == Ctrl-Up on Eterm. */
-			kbinput = NANO_UP_KEY;
-			break;
 		    case 'b': /* Esc o b == Ctrl-Down on Eterm. */
-			kbinput = NANO_DOWN_KEY;
-			break;
 		    case 'c': /* Esc o c == Ctrl-Right on Eterm. */
-			kbinput = NANO_FORWARD_KEY;
-			break;
 		    case 'd': /* Esc o d == Ctrl-Left on Eterm. */
-			kbinput = NANO_BACK_KEY;
+			kbinput = get_escape_seq_abcd(escape_seq[1]);
 			break;
 		}
 		break;
 	    case '[':
 		switch (escape_seq[1]) {
 		    case '1':
-			if (escape_seq_len >= 5) {
-			    if (!strncmp(escape_seq, "[1;2", 4)) {
-				switch (escape_seq[4]) {
-				    case 'A':
-					/* Esc [ 1 ; 2 A == Shift-Up on
-					 * xterm. */
-					kbinput = NANO_UP_KEY;
-					break;
-				    case 'B':
-					/* Esc [ 1 ; 2 B == Shift-Down
-					 * on xterm. */
-					kbinput = NANO_DOWN_KEY;
-					break;
-				    case 'C':
-					/* Esc [ 1 ; 2 C == Shift-Right
-					 * on xterm. */
-					kbinput = NANO_FORWARD_KEY;
-					break;
-				    case 'D':
-					/* Esc [ 1 ; 2 D == Shift-Left
-					 * on xterm. */
-					kbinput = NANO_BACK_KEY;
-					break;
-				}
-			    } else if (!strncmp(escape_seq, "[1;5", 4)) {
-				switch (escape_seq[4]) {
-				    case 'A':
-					/* Esc [ 1 ; 5 A == Ctrl-Up on
-					 * xterm. */
-					kbinput = NANO_UP_KEY;
-					break;
-				    case 'B':
-					/* Esc [ 1 ; 5 B == Ctrl-Down on
-					 * xterm. */
-					kbinput = NANO_DOWN_KEY;
-					break;
-				    case 'C':
-					/* Esc [ 1 ; 5 C == Ctrl-Right
-					 * on xterm. */
-					kbinput = NANO_FORWARD_KEY;
-					break;
-				    case 'D':
-					/* Esc [ 1 ; 5 D == Ctrl-Left on
-					 * xterm. */
-					kbinput = NANO_BACK_KEY;
-					break;
-				}
+			if (escape_seq_len >= 3) {
+			    switch (escape_seq[2]) {
+				case '1': /* Esc [ 1 1 ~ == F1 on
+					   * rxvt/Eterm. */
+				    kbinput = KEY_F(1);
+				    break;
+				case '2': /* Esc [ 1 2 ~ == F2 on
+					   * rxvt/Eterm. */
+				    kbinput = KEY_F(2);
+				    break;
+				case '3': /* Esc [ 1 3 ~ == F3 on
+					   * rxvt/Eterm. */
+				    kbinput = KEY_F(3);
+				    break;
+				case '4': /* Esc [ 1 4 ~ == F4 on
+					   * rxvt/Eterm. */
+				    kbinput = KEY_F(4);
+				    break;
+				case '5': /* Esc [ 1 5 ~ == F5 on
+					   * xterm/rxvt/Eterm. */
+				    kbinput = KEY_F(5);
+				    break;
+				case '7': /* Esc [ 1 7 ~ == F6 on Linux
+					   * console/xterm/rxvt/Eterm. */
+				    kbinput = KEY_F(6);
+				    break;
+				case '8': /* Esc [ 1 8 ~ == F7 on Linux
+					   * console/xterm/rxvt/Eterm. */
+				    kbinput = KEY_F(7);
+				    break;
+				case '9': /* Esc [ 1 9 ~ == F8 on Linux
+					   * console/xterm/rxvt/Eterm. */
+				    kbinput = KEY_F(8);
+				    break;
+				case ';':
+    if (escape_seq_len >= 4) {
+	switch (escape_seq[3]) {
+	    case '2':
+		if (escape_seq_len >= 5) {
+		    switch (escape_seq[4]) {
+			case 'A': /* Esc [ 1 ; 2 A == Shift-Up on
+				   * xterm. */
+			case 'B': /* Esc [ 1 ; 2 B == Shift-Down on
+				   * xterm. */
+			case 'C': /* Esc [ 1 ; 2 C == Shift-Right on
+				   * xterm. */
+			case 'D': /* Esc [ 1 ; 2 D == Shift-Left on
+				   * xterm. */
+			    kbinput = get_escape_seq_abcd(escape_seq[4]);
+			    break;
+		    }
+		}
+		break;
+	    case '5':
+		if (escape_seq_len >= 5) {
+		    switch (escape_seq[4]) {
+			case 'A': /* Esc [ 1 ; 5 A == Ctrl-Up on
+				   * xterm. */
+			case 'B': /* Esc [ 1 ; 5 B == Ctrl-Down on
+				   * xterm. */
+			case 'C': /* Esc [ 1 ; 5 C == Ctrl-Right on
+				   * xterm. */
+			case 'D': /* Esc [ 1 ; 5 D == Ctrl-Left on
+				   * xterm. */
+			    kbinput = get_escape_seq_abcd(escape_seq[4]);
+			    break;
+		    }
+		}
+		break;
+	}
+    }
+				    break;
+				default: /* Esc [ 1 ~ == Home on Linux
+					  * console. */
+				    kbinput = NANO_HOME_KEY;
+				    break;
+			    }
+			}
+			break;
+		    case '2':
+			if (escape_seq_len >= 3) {
+			    switch (escape_seq[2]) {
+				case '0': /* Esc [ 2 0 ~ == F9 on Linux
+					   * console/xterm/rxvt/Eterm. */
+				    kbinput = KEY_F(9);
+				    break;
+				case '1': /* Esc [ 2 1 ~ == F10 on Linux
+					   * console/xterm/rxvt/Eterm. */
+				    kbinput = KEY_F(10);
+				    break;
+				case '3': /* Esc [ 2 3 ~ == F11 on Linux
+					   * console/xterm/rxvt/Eterm. */
+				    kbinput = KEY_F(11);
+				    break;
+				case '4': /* Esc [ 2 4 ~ == F12 on Linux
+					   * console/xterm/rxvt/Eterm. */
+				    kbinput = KEY_F(12);
+				    break;
+				case '5': /* Esc [ 2 5 ~ == F13 on Linux
+					   * console/rxvt/Eterm. */
+				    kbinput = KEY_F(13);
+				    break;
+				case '6': /* Esc [ 2 6 ~ == F14 on Linux
+					   * console/rxvt/Eterm. */
+				    kbinput = KEY_F(14);
+				    break;
+				default: /* Esc [ 2 ~ == Insert on Linux
+					  * console/xterm. */
+				    kbinput = NANO_INSERTFILE_KEY;
+				    break;
 			    }
-			    break;
-			} else {
-			    /* Esc [ 1 ~ == Home on Linux console. */
-			    kbinput = NANO_HOME_KEY;
-			    break;
 			}
-		    case '2': /* Esc [ 2 ~ == Insert on Linux
-			       * console/xterm. */
-			kbinput = NANO_INSERTFILE_KEY;
 			break;
 		    case '3': /* Esc [ 3 ~ == Delete on Linux
 			       * console/xterm. */
@@ -432,12 +576,12 @@ int get_escape_seq_kbinput(WINDOW *win, char *escape_seq, int
 			kbinput = NANO_END_KEY;
 			break;
 		    case '5': /* Esc [ 5 ~ == PageUp on Linux
-			       * console/xterm, Esc [ 5 ^ == PageUp on
+			       * console/xterm; Esc [ 5 ^ == PageUp on
 			       * Eterm. */
 			kbinput = NANO_PREVPAGE_KEY;
 			break;
 		    case '6': /* Esc [ 6 ~ == PageDown on Linux
-			       * console/xterm, Esc [ 6 ^ == PageDown on
+			       * console/xterm; Esc [ 6 ^ == PageDown on
 			       * Eterm. */
 			kbinput = NANO_NEXTPAGE_KEY;
 			break;
@@ -455,22 +599,16 @@ int get_escape_seq_kbinput(WINDOW *win, char *escape_seq, int
 			break;
 		    case 'A': /* Esc [ A == Up on Linux console/FreeBSD
 			       * console/Hurd console/rxvt/Eterm. */
-			kbinput = NANO_UP_KEY;
-			break;
 		    case 'B': /* Esc [ B == Down on Linux
 			       * console/FreeBSD console/Hurd
 			       * console/rxvt/Eterm. */
-			kbinput = NANO_DOWN_KEY;
-			break;
 		    case 'C': /* Esc [ C == Right on Linux
 			       * console/FreeBSD console/Hurd
 			       * console/rxvt/Eterm. */
-			kbinput = NANO_FORWARD_KEY;
-			break;
 		    case 'D': /* Esc [ D == Left on Linux
 			       * console/FreeBSD console/Hurd
 			       * console/rxvt/Eterm. */
-			kbinput = NANO_BACK_KEY;
+			kbinput = get_escape_seq_abcd(escape_seq[1]);
 			break;
 		    case 'F': /* Esc [ F == End on FreeBSD
 			       * console/Eterm. */
@@ -492,34 +630,110 @@ int get_escape_seq_kbinput(WINDOW *win, char *escape_seq, int
 			       * console. */
 			kbinput = NANO_INSERTFILE_KEY;
 			break;
+		    case 'M': /* Esc [ M == F1 on FreeBSD console. */
+			kbinput = KEY_F(1);
+			break;
+		    case 'N': /* Esc [ N == F2 on FreeBSD console. */
+			kbinput = KEY_F(2);
+			break;
+		    case 'O':
+			if (escape_seq_len >= 3) {
+			    switch (escape_seq[2]) {
+				case 'P': /* Esc [ O P == F1 on
+					   * xterm. */
+				    kbinput = KEY_F(1);
+				    break;
+				case 'Q': /* Esc [ O Q == F2 on
+					   * xterm. */
+				    kbinput = KEY_F(2);
+				    break;
+				case 'R': /* Esc [ O R == F3 on
+					   * xterm. */
+				    kbinput = KEY_F(3);
+				    break;
+				case 'S': /* Esc [ O S == F4 on
+					   * xterm. */
+				    kbinput = KEY_F(4);
+				    break;
+				default: /* Esc [ O == F3 on
+					  * FreeBSD console. */
+				    kbinput = KEY_F(3);
+				    break;
+			    }
+			}
+			break;
+		    case 'P': /* Esc [ P == F4 on FreeBSD console. */
+			kbinput = KEY_F(4);
+			break;
+		    case 'Q': /* Esc [ Q == F5 on FreeBSD console. */
+			kbinput = KEY_F(5);
+			break;
+		    case 'R': /* Esc [ R == F6 on FreeBSD console. */
+			kbinput = KEY_F(6);
+			break;
+		    case 'S': /* Esc [ S == F7 on FreeBSD console. */
+			kbinput = KEY_F(7);
+			break;
+		    case 'T': /* Esc [ T == F8 on FreeBSD console. */
+			kbinput = KEY_F(8);
+			break;
 		    case 'U': /* Esc [ U == PageDown on Hurd console. */
 			kbinput = NANO_NEXTPAGE_KEY;
 			break;
 		    case 'V': /* Esc [ V == PageUp on Hurd console. */
 			kbinput = NANO_PREVPAGE_KEY;
 			break;
+		    case 'W': /* Esc [ W == F11 on FreeBSD console. */
+			kbinput = KEY_F(11);
+			break;
+		    case 'X': /* Esc [ X == F12 on FreeBSD console. */
+			kbinput = KEY_F(12);
+			break;
 		    case 'Y': /* Esc [ Y == End on Hurd console. */
 			kbinput = NANO_END_KEY;
 			break;
+		    case 'Z': /* Esc [ Z == F14 on FreeBSD console. */
+			kbinput = KEY_F(14);
+			break;
 		    case 'a': /* Esc [ a == Shift-Up on rxvt/Eterm. */
-			kbinput = NANO_UP_KEY;
-			break;
 		    case 'b': /* Esc [ b == Shift-Down on rxvt/Eterm. */
-			kbinput = NANO_DOWN_KEY;
-			break;
 		    case 'c': /* Esc [ c == Shift-Right on
 			       * rxvt/Eterm. */
-			kbinput = NANO_FORWARD_KEY;
-			break;
 		    case 'd': /* Esc [ d == Shift-Left on rxvt/Eterm. */
-			kbinput = NANO_BACK_KEY;
+			kbinput = get_escape_seq_abcd(escape_seq[1]);
+			break;
+		    case '[':
+			if (escape_seq_len >= 3) {
+			    switch (escape_seq[2]) {
+				case 'A': /* Esc [ [ A == F1 on Linux
+					   * console. */
+				    kbinput = KEY_F(1);
+				    break;
+				case 'B': /* Esc [ [ B == F2 on Linux
+					   * console. */
+				    kbinput = KEY_F(2);
+				    break;
+				case 'C': /* Esc [ [ C == F3 on Linux
+					   * console. */
+				    kbinput = KEY_F(3);
+				    break;
+				case 'D': /* Esc [ [ D == F4 on Linux
+					   * console. */
+				    kbinput = KEY_F(4);
+				    break;
+				case 'E': /* Esc [ [ E == F5 on Linux
+					   * console. */
+				    kbinput = KEY_F(5);
+				    break;
+			    }
+			}
 			break;
 		}
 		break;
 	}
     }
 
-    if (kbinput == -1) {
+    if (kbinput == ERR) {
 	/* This escape sequence is unrecognized; send it back. */
 	for (; escape_seq_len > 1; escape_seq_len--)
 	    ungetch(escape_seq[escape_seq_len - 1]);
@@ -529,6 +743,25 @@ int get_escape_seq_kbinput(WINDOW *win, char *escape_seq, int
     return kbinput;
 }
 
+/* Return the equivalent arrow key value for the case-insensitive
+ * letters A (up), B (down), C (right), and D (left).  These are common
+ * to many escape sequences. */
+int get_escape_seq_abcd(int kbinput)
+{
+    switch (tolower(kbinput)) {
+	case 'a':
+	    return NANO_PREVLINE_KEY;
+	case 'b':
+	    return NANO_NEXTLINE_KEY;
+	case 'c':
+	    return NANO_FORWARD_KEY;
+	case 'd':
+	    return NANO_BACK_KEY;
+	default:
+	    return ERR;
+    }
+}
+
 #ifndef DISABLE_MOUSE
 /* Check for a mouse event, and if one's taken place, save the
  * coordinates where it took place in mouse_x and mouse_y.  After that,
@@ -600,13 +833,16 @@ int get_mouseinput(int *mouse_x, int *mouse_y, int allow_shortcuts)
 	for (; j > 0; j--)
 	    s = s->next;
 
-	/* And ungetch() the equivalent keystroke. */
-	ungetch(s->val);
-
-	/* If it's not a control character, assume it's a Meta key
-	 * sequence, in which case we need to ungetch() Escape too. */
-	if (!is_cntrl_char(s->val))
-	   ungetch(NANO_CONTROL_3);
+	/* And ungetch() the equivalent control key.  If it's a Meta key
+	 * sequence, we need to ungetch() Escape too.  Assume that the
+	 * shortcut has an equivalent control key, meta key sequence, or
+	 * both. */
+	if (s->ctrlval != NANO_NO_KEY)
+	    ungetch(s->ctrlval);
+	else {
+	    ungetch(s->metaval);
+	    ungetch(NANO_CONTROL_3);
+	}
 
 	return 1;
     }
@@ -690,7 +926,7 @@ size_t strnlenpt(const char *buf, size_t size)
 /* How many columns wide is buf? */
 size_t strlenpt(const char *buf)
 {
-    return strnlenpt(buf, -1);
+    return strnlenpt(buf, (size_t)-1);
 }
 
 void blank_bottombars(void)
@@ -832,7 +1068,7 @@ void nanoget_repaint(const char *buf, const char *inputbuf, size_t x)
     wattroff(bottomwin, A_REVERSE);
 }
 
-/* Get the input from the kb; this should only be called from
+/* Get the input from the keyboard; this should only be called from
  * statusq(). */
 int nanogetstr(int allowtabs, const char *buf, const char *def,
 #ifndef NANO_SMALL
@@ -845,7 +1081,7 @@ int nanogetstr(int allowtabs, const char *buf, const char *def,
 		)
 {
     int kbinput;
-    int meta;
+    int meta_key;
     static int x = -1;
 	/* the cursor position in 'answer' */
     int xend;
@@ -896,29 +1132,33 @@ int nanogetstr(int allowtabs, const char *buf, const char *def,
        input */
     wrefresh(edit);
 
-    while ((kbinput = get_kbinput(bottomwin, &meta)) != NANO_ENTER_KEY) {
+    while ((kbinput = get_kbinput(bottomwin, &meta_key)) != NANO_ENTER_KEY) {
 	for (t = s; t != NULL; t = t->next) {
 #ifdef DEBUG
 	    fprintf(stderr, "Aha! \'%c\' (%d)\n", kbinput, kbinput);
 #endif
 
-	    if (kbinput == t->val && is_cntrl_char(kbinput)) {
+	    /* Temporary hack to interpret NANO_HELP_FKEY correctly. */
+	    if (kbinput == t->funcval)
+		kbinput = t->ctrlval;
+
+	    if (kbinput == t->ctrlval && is_cntrl_char(kbinput)) {
 
 #ifndef DISABLE_HELP
 		/* Have to do this here, it would be too late to do it
 		   in statusq() */
-		if (kbinput == NANO_HELP_KEY || kbinput == NANO_HELP_FKEY) {
+		if (kbinput == NANO_HELP_KEY) {
 		    do_help();
 		    break;
 		}
 #endif
 #ifndef NANO_SMALL
 		/* Have to handle these here too, for the time being */
-		if (kbinput == NANO_UP_KEY || kbinput == NANO_DOWN_KEY)
+		if (kbinput == NANO_PREVLINE_KEY || kbinput == NANO_NEXTLINE_KEY)
 		    break;
 #endif
 
-		return t->val;
+		return t->ctrlval;
 	    }
 	}
 	assert(0 <= x && x <= xend && xend == strlen(answer));
@@ -997,7 +1237,7 @@ int nanogetstr(int allowtabs, const char *buf, const char *def,
 	    if (x > 0)
 		x--;
 	    break;
-	case NANO_UP_KEY:
+	case NANO_PREVLINE_KEY:
 #ifndef NANO_SMALL
 	    if (history_list != NULL) {
 
@@ -1039,7 +1279,7 @@ int nanogetstr(int allowtabs, const char *buf, const char *def,
 	    }
 #endif
 	    break;
-	case NANO_DOWN_KEY:
+	case NANO_NEXTLINE_KEY:
 #ifndef NANO_SMALL
 	    if (history_list != NULL) {
 
@@ -1089,7 +1329,7 @@ int nanogetstr(int allowtabs, const char *buf, const char *def,
 		    fprintf(stderr, "Aha! \'%c\' (%d)\n", kbinput,
 			    kbinput);
 #endif
-		    if (meta == 1 && (kbinput == t->metaval || kbinput == t->misc))
+		    if (meta_key == TRUE && (kbinput == t->metaval || kbinput == t->miscval))
 			/* We hit a Meta key.  Do like above.  We don't
 			 * just ungetch() the letter and let it get
 			 * caught above cause that screws the
@@ -1210,18 +1450,18 @@ void bottombars(const shortcut *s)
 	    wmove(bottomwin, 1 + j, i * (COLS / numcols));
 
 	    /* Yucky sentinel values we can't handle a better way */
-	    if (s->val != NANO_NO_KEY) {
+	    if (s->ctrlval != NANO_NO_KEY) {
 #ifndef NANO_SMALL
-		if (s->val == NANO_HISTORY_KEY)
+		if (s->ctrlval == NANO_HISTORY_KEY)
 		    strncpy(keystr, _("Up"), 8);
 		else
 #endif
-		if (s->val == NANO_CONTROL_SPACE)
+		if (s->ctrlval == NANO_CONTROL_SPACE)
 		    strcpy(keystr, "^ ");
-		else if (s->val == NANO_CONTROL_8)
+		else if (s->ctrlval == NANO_CONTROL_8)
 		    strcpy(keystr, "^?");
 		else
-		    sprintf(keystr, "^%c", s->val + 64);
+		    sprintf(keystr, "^%c", s->ctrlval + 64);
 	    } else if (s->metaval != NANO_NO_KEY)
 		sprintf(keystr, "M-%c", toupper(s->metaval));
 
@@ -1281,7 +1521,7 @@ size_t get_page_start(size_t column)
     if (column == 0 || column < COLS - 1)
 	return 0;
     else if (COLS > 9)
-	return column - 7 - (column - 8) % (COLS - 9);
+	return column - 7 - (column - 7) % (COLS - 8);
     else if (COLS > 2)
 	return column - (COLS - 2);
     else
@@ -1293,23 +1533,17 @@ size_t get_page_start(size_t column)
  * cursor at (current_y, current_x). */
 void reset_cursor(void)
 {
-    const filestruct *ptr = edittop;
-    size_t x;
-
     /* Yuck.  This condition can be true after open_file() when opening
      * the first file. */
     if (edittop == NULL)
 	return;
 
-    current_y = 0;
+    current_y = current->lineno - edittop->lineno;
+    if (current_y < editwinrows) {
+	size_t x = xplustabs();
 
-    while (ptr != current && ptr != editbot && ptr->next != NULL) {
-	ptr = ptr->next;
-	current_y++;
-    }
-
-    x = xplustabs();
-    wmove(edit, current_y, x - get_page_start(x));
+	wmove(edit, current_y, x - get_page_start(x));
+     }
 }
 
 /* edit_add() takes care of the job of actually painting a line into the
@@ -1441,7 +1675,7 @@ void edit_add(const filestruct *fileptr, const char *converted,
 		 * line not followed by an end on this line? */
 
 		start_col = 0;
-		while (1) {
+		while (TRUE) {
 		    start_col += startmatch.rm_so;
 		    startmatch.rm_eo -= startmatch.rm_so;
  		    if (regexec(tmpcolor->end,
@@ -1718,26 +1952,33 @@ void edit_refresh(void)
 	edit_update(current, CENTER);
     else {
 	int nlines = 0;
+	const filestruct *foo = edittop;
 
-	/* Don't make the cursor jump around the screen whilst
+#ifdef DEBUG
+	fprintf(stderr, "edit_refresh(): edittop->lineno = %ld\n", edittop->lineno);
+#endif
+
+	/* Don't let the cursor jump around the screen while
 	 * updating. */
 	leaveok(edit, TRUE);
 
-	editbot = edittop;
 	while (nlines < editwinrows) {
-	    update_line(editbot, current_x);
+	    update_line(foo, current_x);
 	    nlines++;
-	    if (editbot->next == NULL)
+	    if (foo->next == NULL)
 		break;
-	    editbot = editbot->next;
+	    foo = foo->next;
 	}
 	while (nlines < editwinrows) {
 	    mvwaddstr(edit, nlines, 0, hblank);
 	    nlines++;
 	}
+	reset_cursor();
 	/* What the hell are we expecting to update the screen if this
 	 * isn't here?  Luck? */
 	wrefresh(edit);
+
+	/* Let the cursor jump around the screen again. */
 	leaveok(edit, FALSE);
     }
 }
@@ -1774,7 +2015,7 @@ void edit_update(filestruct *fileptr, topmidnone location)
  * want to put up by default.
  *
  * New arg tabs tells whether or not to allow tab completion. */
-int statusq(int tabs, const shortcut *s, const char *def,
+int statusq(int allowtabs, const shortcut *s, const char *def,
 #ifndef NANO_SMALL
 		historyheadtype *which_history,
 #endif
@@ -1794,7 +2035,7 @@ int statusq(int tabs, const shortcut *s, const char *def,
     va_end(ap);
     foo[COLS - 4] = '\0';
 
-    ret = nanogetstr(tabs, foo, def,
+    ret = nanogetstr(allowtabs, foo, def,
 #ifndef NANO_SMALL
 		which_history,
 #endif
@@ -1808,10 +2049,12 @@ int statusq(int tabs, const shortcut *s, const char *def,
 
     switch (ret) {
     case NANO_FIRSTLINE_KEY:
+    case NANO_FIRSTLINE_FKEY:
 	do_first_line();
 	resetstatuspos = 1;
 	break;
     case NANO_LASTLINE_KEY:
+    case NANO_LASTLINE_FKEY:
 	do_last_line();
 	resetstatuspos = 1;
 	break;
@@ -1904,12 +2147,12 @@ int do_yesno(int all, const char *msg)
 
     do {
 	int kbinput;
-	int meta;
+	int meta_key;
 #ifndef DISABLE_MOUSE
 	int mouse_x, mouse_y;
 #endif
 
-	kbinput = get_kbinput(edit, &meta);
+	kbinput = get_kbinput(edit, &meta_key);
 
 	if (kbinput == NANO_CANCEL_KEY)
 	    ok = -1;
@@ -2116,7 +2359,7 @@ int line_len(const char *ptr)
 int do_help(void)
 {
 #ifndef DISABLE_HELP
-    int i, page = 0, kbinput = -1, meta, no_more = 0;
+    int i, page = 0, kbinput = ERR, meta_key, no_more = 0;
     int no_help_flag = 0;
     const shortcut *oldshortcut;
 
@@ -2192,7 +2435,7 @@ int do_help(void)
 	    no_more = 1;
 	    continue;
 	}
-    } while ((kbinput = get_kbinput(edit, &meta)) != NANO_EXIT_KEY && kbinput != NANO_EXIT_FKEY);
+    } while ((kbinput = get_kbinput(edit, &meta_key)) != NANO_EXIT_KEY && kbinput != NANO_EXIT_FKEY);
 
     currshortcut = oldshortcut;
 
@@ -2252,16 +2495,6 @@ void do_replace_highlight(int highlight_flag, const char *word)
 	wattroff(edit, A_REVERSE);
 }
 
-/* Fix editbot, based on the assumption that edittop is correct. */
-void fix_editbot(void)
-{
-    int i;
-
-    editbot = edittop;
-    for (i = 0; i < editwinrows && editbot->next != NULL; i++)
-	editbot = editbot->next;
-}
-
 #ifdef DEBUG
 /* Dump the passed-in file structure to stderr. */
 void dump_buffer(const filestruct *inptr)
@@ -2295,9 +2528,11 @@ void dump_buffer_reverse(void)
 #define CREDIT_LEN 53
 #define XLCREDIT_LEN 8
 
+/* Easter egg: Display credits.  Assume nodelay(edit) is FALSE. */
 void do_credits(void)
 {
     int i, j = 0, k, place = 0, start_x;
+    struct timespec scrolldelay;
 
     const char *what;
     const char *xlcredits[XLCREDIT_LEN];
@@ -2361,6 +2596,9 @@ void do_credits(void)
     xlcredits[6] = _("and anyone else we forgot...");
     xlcredits[7] = _("Thank you for using nano!\n");
 
+    scrolldelay.tv_sec = 0;
+    scrolldelay.tv_nsec = 700000000;
+
     curs_set(0);
     nodelay(edit, TRUE);
     blank_bottombars();
@@ -2392,7 +2630,7 @@ void do_credits(void)
 		start_x = COLS / 2 - strlen(what) / 2 - 1;
 		mvwaddstr(edit, i * 2 - k, start_x, what);
 	    }
-	    usleep(700000);
+	    nanosleep(&scrolldelay, NULL);
 	    wrefresh(edit);
 	}
 	if (j < editwinrows / 2 - 1)
